!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-------------------------------------------------------------------------------
MODULE MODI_ICE4_FAST_RI
INTERFACE
SUBROUTINE ICE4_FAST_RI(LDSOFT, PCOMPUTE, &
                       &PRHODREF, PLVFACT, PLSFACT, &
                       &PAI, PCJ, PCIT, &
                       &PSSI, &
                       &PRCT, PRIT, &
                       &PRCBERI, PA_TH, PA_RC, PA_RI)
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_RAIN_ICE_PARAM
USE MODD_RAIN_ICE_DESCR
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PAI      ! Thermodynamical function
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PSSI     ! Supersaturation over ice
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCBERI  ! Bergeron-Findeisen effect
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
END SUBROUTINE ICE4_FAST_RI
END INTERFACE
END MODULE MODI_ICE4_FAST_RI
SUBROUTINE ICE4_FAST_RI(LDSOFT, PCOMPUTE, &
                       &PRHODREF, PLVFACT, PLSFACT, &
                       &PAI, PCJ, PCIT, &
                       &PSSI, &
                       &PRCT, PRIT, &
                       &PRCBERI, PA_TH, PA_RC, PA_RI)
!!
!!**  PURPOSE
!!    -------
!!      Computes the fast ri process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_RAIN_ICE_DESCR, ONLY: XDI,XLBEXI,XLBI,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: X0DEPI,X2DEPI
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PAI      ! Thermodynamical function
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PSSI     ! Supersaturation over ice
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCBERI  ! Bergeron-Findeisen effect
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
!
!*       0.2  declaration of local variables
!
INTEGER                            :: ISIZE
INTEGER                            :: JL
#ifndef MNH_OPENACC
REAL,    DIMENSION(:), allocatable :: ZMASK
#else
REAL,    DIMENSION(:), pointer, contiguous :: ZMASK
#endif
!
!-------------------------------------------------------------------------------
!
!$acc data present(PCOMPUTE,PRHODREF,PLVFACT,PLSFACT,PAI,PCJ,PCIT, &
!$acc&                PSSI,PRCT,PRIT,PRCBERI,PA_TH,PA_RC,PA_RI)

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PCOMPUTE,"ICE4_FAST_RI beg:PCOMPUTE")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_FAST_RI beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_FAST_RI beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PLSFACT,"ICE4_FAST_RI beg:PLSFACT",PRECISION)
  CALL MPPDB_CHECK1D(PAI,"ICE4_FAST_RI beg:PAI",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_FAST_RI beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PCIT,"ICE4_FAST_RI beg:PCIT",PRECISION)
  CALL MPPDB_CHECK1D(PSSI,"ICE4_FAST_RI beg:PSSI",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_FAST_RI beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRIT,"ICE4_FAST_RI beg:PRIT",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRCBERI,"ICE4_FAST_RI beg:PRCBERI",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RI beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RI beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RI beg:PA_RI",PRECISION)
END IF

ISIZE = Size( PRHODREF )

#ifndef MNH_OPENACC
allocate( zmask( size( prhodref ) ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_FAST_RI' )

CALL MNH_MEM_GET( zmask, ISIZE )

!$acc data create(ZMASK)
#endif

!
!*       7.2    Bergeron-Findeisen effect: RCBERI
!
!$acc kernels
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., -PSSI(JL))) * &          ! PSSI(:)>0.
           &MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! PRCT(:)>XRTMIN(2)
           &MAX(0., -SIGN(1., XRTMIN(4)-PRIT(JL))) * & ! PRIT(:)>XRTMIN(4)
           &MAX(0., -SIGN(1., 1.E-20-PCIT(JL))) * &          ! PCIT(:)>0.
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRCBERI(JL) = PRCBERI(JL) * ZMASK(JL)
  ENDDO
ELSE
  PRCBERI(:) = 0.
  WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
    PRCBERI(:) = MIN(1.E8, XLBI*(PRHODREF(:)*PRIT(:)/PCIT(:))**XLBEXI) ! Lbda_i
    PRCBERI(:) = ( PSSI(:) / (PRHODREF(:)*PAI(:)) ) * PCIT(:) * &
                 ( X0DEPI/PRCBERI(:) + X2DEPI*PCJ(:)*PCJ(:)/PRCBERI(:)**(XDI+2.0) )
#else
    PRCBERI(:) = MIN(1.E8, XLBI*BR_POW(PRHODREF(:)*PRIT(:)/PCIT(:),XLBEXI)) ! Lbda_i
    PRCBERI(:) = ( PSSI(:) / (PRHODREF(:)*PAI(:)) ) * PCIT(:) * &
                 ( X0DEPI/PRCBERI(:) + X2DEPI*PCJ(:)*PCJ(:)/BR_POW(PRCBERI(:),XDI+2.0) )
#endif
  END WHERE
ENDIF

DO JL=1, ISIZE
  PA_RC(JL) = PA_RC(JL) - PRCBERI(JL)
  PA_RI(JL) = PA_RI(JL) + PRCBERI(JL)
  PA_TH(JL) = PA_TH(JL) + PRCBERI(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRCBERI,"ICE4_FAST_RI end:PRCBERI",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RI end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RI end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RI end:PA_RI",PRECISION)
END IF

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'ICE4_FAST_RI' )
#endif

!$acc end data

END SUBROUTINE ICE4_FAST_RI
