!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #######################
     MODULE MODI_TURB_HOR_UV
!    #######################
!
INTERFACE  
!
      SUBROUTINE TURB_HOR_UV(KSPLT,                                  &
                      OTURB_FLX,                                     &
                      TPFILE,                                        &
                      PK,PINV_PDXX,PINV_PDYY,PINV_PDZZ,PMZM_PRHODJ,  &
                      PDXX,PDYY,PDZZ,PDZX,PDZY,                      &
                      PDIRCOSZW,                                     &
                      PCOSSLOPE,PSINSLOPE,                           &
                      PRHODJ,                                        &
                      PCDUEFF,PTAU11M,PTAU12M,PTAU22M,PTAU33M,       &
                      PUM,PVM,PUSLOPEM,PVSLOPEM,                     &
                      PDP,                                           &
                      PRUS,PRVS                                      )
!
USE MODD_IO, ONLY: TFILEDATA
!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
TYPE(TFILEDATA),          INTENT(INOUT) ::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDXX   ! 1./PDXX
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDYY   ! 1./PDYY
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDXX, PDYY, PDZZ, PDZX, PDZY 
                                                         ! Metric coefficients
REAL, DIMENSION(:,:),     INTENT(IN)    ::  PDIRCOSZW
! Director Cosinus along z directions at surface w-point
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE       ! cosinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE       ! sinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCDUEFF      ! Cd * || u || at time t
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU11M      ! <uu> in the axes linked 
       ! to the maximum slope direction and the surface normal and the binormal 
       ! at time t - dt
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU12M      ! <uv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU22M      ! <vv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU33M      ! <ww> in the same axes
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PUM,PVM
REAL, DIMENSION(:,:),      INTENT(IN)   ::  PUSLOPEM     ! wind component along the 
                                     ! maximum slope direction
REAL, DIMENSION(:,:),      INTENT(IN)   ::  PVSLOPEM     ! wind component along the 
                                     ! direction normal to the maximum slope one
!
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRUS, PRVS   ! var. at t+1 -split-
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!
END SUBROUTINE TURB_HOR_UV
!
END INTERFACE
!
END MODULE MODI_TURB_HOR_UV
!     ################################################################
      SUBROUTINE TURB_HOR_UV(KSPLT,                                  &
                      OTURB_FLX,                                     &
                      TPFILE,                                        &
                      PK,PINV_PDXX,PINV_PDYY,PINV_PDZZ,PMZM_PRHODJ,  &
                      PDXX,PDYY,PDZZ,PDZX,PDZY,                      &
                      PDIRCOSZW,                                     &
                      PCOSSLOPE,PSINSLOPE,                           &
                      PRHODJ,                                        &
                      PCDUEFF,PTAU11M,PTAU12M,PTAU22M,PTAU33M,       &
                      PUM,PVM,PUSLOPEM,PVSLOPEM,                     &
                      PDP,                                           &
                      PRUS,PRVS                                      )
!     ################################################################
!
!
!!****  *TURB_HOR* -routine to compute the source terms in the meso-NH
!!               model equations due to the non-vertical turbulent fluxes.
!!
!!    PURPOSE
!!    -------
!!
!!     see TURB_HOR
!!
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!                     Aug    , 1997 (V. Saravane) spliting of TURB_HOR
!!                     Nov  27, 1997 (V. Masson) clearing of the routine
!!                     Oct  18, 2000 (V. Masson) LES computations + LFLAT switch
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!!  Philippe Wautelet: 05/2016-04/2018: new data structures and calls for I/O
!! --------------------------------------------------------------------------
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_CONF
USE MODD_CTURB
#ifdef MNH_OPENACC
USE MODD_DIM_n,          only: NKMAX
#endif
use modd_field,          only: tfielddata, TYPEREAL
USE MODD_IO,             ONLY: TFILEDATA
USE MODD_PARAMETERS
USE MODD_LES
!
USE MODE_IO_FIELD_WRITE, only: IO_Field_write
use mode_mppdb
!
USE MODI_GRADIENT_M
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_COEFJ
USE MODI_LES_MEAN_SUBGRID
!
USE MODI_SECOND_MNH
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
!
IMPLICIT NONE
!
!
!*       0.1  declaration of arguments
!
!
!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
TYPE(TFILEDATA),          INTENT(INOUT) ::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDXX   ! 1./PDXX
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDYY   ! 1./PDYY
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDXX, PDYY, PDZZ, PDZX, PDZY 
                                                         ! Metric coefficients
REAL, DIMENSION(:,:),     INTENT(IN)    ::  PDIRCOSZW
! Director Cosinus along z directions at surface w-point
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE       ! cosinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE       ! sinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCDUEFF      ! Cd * || u || at time t
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU11M      ! <uu> in the axes linked 
       ! to the maximum slope direction and the surface normal and the binormal 
       ! at time t - dt
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU12M      ! <uv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU22M      ! <vv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU33M      ! <ww> in the same axes
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PUM,PVM
REAL, DIMENSION(:,:),      INTENT(IN)   ::  PUSLOPEM     ! wind component along the 
                                     ! maximum slope direction
REAL, DIMENSION(:,:),      INTENT(IN)   ::  PVSLOPEM     ! wind component along the 
                                     ! direction normal to the maximum slope one
!
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRUS, PRVS   ! var. at t+1 -split-
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!
!*       0.2  declaration of local variables
!
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZFLX,ZWORK ! work arrays
!
REAL, DIMENSION(:,:),   pointer , contiguous :: ZDIRSINZW
      ! sinus of the angle between the vertical and the normal to the orography
INTEGER             :: IKB,IKE
                                    ! Index values for the Beginning and End
                                    ! mass points of the domain  
!
REAL, DIMENSION(:,:,:), pointer , contiguous :: GY_U_UV_PUM
REAL, DIMENSION(:,:,:), pointer , contiguous :: GX_V_UV_PVM
!
REAL :: ZTIME1, ZTIME2
!
#ifdef MNH_OPENACC
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP1_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP2_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP3_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP4_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP5_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP6_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP7_DEVICE
#endif
TYPE(TFIELDDATA) :: TZFIELD
!
INTEGER  :: JIU,JJU,JKU
INTEGER  :: JI,JJ,JK
! ---------------------------------------------------------------------------

!$acc data present( PK, PINV_PDXX, PINV_PDYY, PINV_PDZZ, PMZM_PRHODJ, PDXX, PDYY, PDZZ, PDZX, PDZY,       &
!$acc &             PDIRCOSZW, PCOSSLOPE, PSINSLOPE, PRHODJ, PCDUEFF, PTAU11M, PTAU12M, PTAU22M, PTAU33M, &
!$acc &             PUM, PVM, PUSLOPEM, PVSLOPEM,                                                         &
!$acc &             PRUS, PRVS, PDP )

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( pk,          "Turb_hor_uv beg:pk"          )
  call Mppdb_check( pinv_pdxx,   "Turb_hor_uv beg:pinv_pdxx"   )
  call Mppdb_check( pinv_pdyy,   "Turb_hor_uv beg:pinv_pdyy"   )
  call Mppdb_check( pinv_pdzz,   "Turb_hor_uv beg:pinv_pdzz"   )
  call Mppdb_check( pmzm_prhodj, "Turb_hor_uv beg:pmzm_prhodj" )
  call Mppdb_check( pdxx,        "Turb_hor_uv beg:pdxx"        )
  call Mppdb_check( pdyy,        "Turb_hor_uv beg:pdyy"        )
  call Mppdb_check( pdzz,        "Turb_hor_uv beg:pdzz"        )
  call Mppdb_check( pdzx,        "Turb_hor_uv beg:pdzx"        )
  call Mppdb_check( pdzy,        "Turb_hor_uv beg:pdzy"        )
  call Mppdb_check( pdircoszw,   "Turb_hor_uv beg:pdircoszw"   )
  call Mppdb_check( pcosslope,   "Turb_hor_uv beg:pcosslope"   )
  call Mppdb_check( psinslope,   "Turb_hor_uv beg:psinslope"   )
  call Mppdb_check( prhodj,      "Turb_hor_uv beg:prhodj"      )
  call Mppdb_check( pcdueff,     "Turb_hor_uv beg:pcdueff"     )
  call Mppdb_check( ptau11m,     "Turb_hor_uv beg:ptau11m"     )
  call Mppdb_check( ptau12m,     "Turb_hor_uv beg:ptau12m"     )
  call Mppdb_check( ptau22m,     "Turb_hor_uv beg:ptau22m"     )
  call Mppdb_check( ptau33m,     "Turb_hor_uv beg:ptau33m"     )
  call Mppdb_check( pum,         "Turb_hor_uv beg:pum"         )
  call Mppdb_check( pvm,         "Turb_hor_uv beg:pvm"         )
  call Mppdb_check( puslopem,    "Turb_hor_uv beg:puslopem"    )
  call Mppdb_check( pvslopem,    "Turb_hor_uv beg:pvslopem"    )
  !Check all inout arrays
  call Mppdb_check( prus,   "Turb_hor_uv beg:prus"   )
  call Mppdb_check( prvs,   "Turb_hor_uv beg:prvs"   )
  call Mppdb_check( pdp,    "Turb_hor_uv beg:pdp"    )
end if

JIU =  size(pum, 1 )
JJU =  size(pum, 2 )
JKU =  size(pum, 3 )

#ifndef MNH_OPENACC
allocate( zflx (JIU,JJU,JKU ) )
allocate( zwork(JIU,JJU,JKU ) )

allocate( zdirsinzw(JIU,JJU ) )

allocate( gy_u_uv_pum(JIU,JJU,JKU ) )
allocate( gx_v_uv_pvm(JIU,JJU,JKU ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( zflx,  JIU, JJU, JKU )
CALL MNH_MEM_GET( zwork, JIU, JJU, JKU )

CALL MNH_MEM_GET( zdirsinzw, JIU, JJU )

CALL MNH_MEM_GET( gy_u_uv_pum, JIU, JJU, JKU )
CALL MNH_MEM_GET( gx_v_uv_pvm, JIU, JJU, JKU )

CALL MNH_MEM_GET( ztmp1_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp2_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp3_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp4_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp5_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp6_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp7_device, JIU, JJU, JKU )
#endif

!$acc data present( ZFLX, ZWORK, ZDIRSINZW, GY_U_UV_PUM, GX_V_UV_PVM,       &
!$acc &            ZTMP1_DEVICE, ZTMP2_DEVICE, ZTMP3_DEVICE, ZTMP4_DEVICE, &
!$acc &            ZTMP5_DEVICE, ZTMP6_DEVICE, ZTMP7_DEVICE                )

!
!*       1.   PRELIMINARY COMPUTATIONS
!             ------------------------
!
IKB = 1+JPVEXT               
IKE = SIZE(PUM,3)-JPVEXT    
!
!$acc kernels
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(2)
#endif
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
#ifndef MNH_BITREP
  ZDIRSINZW(JI,JJ) = SQRT( 1. - PDIRCOSZW(JI,JJ)**2 )
#else
  ZDIRSINZW(JI,JJ) = SQRT( 1. - BR_P2(PDIRCOSZW(JI,JJ)) )
#endif
END DO
!$acc end kernels
!
#ifndef MNH_OPENACC
GX_V_UV_PVM = GX_V_UV(PVM,PDXX,PDZZ,PDZX)
IF (.NOT. L2D) THEN
  GY_U_UV_PUM = GY_U_UV(PUM,PDYY,PDZZ,PDZY)
END IF
#else
CALL GX_V_UV_DEVICE(PVM,PDXX,PDZZ,PDZX,GX_V_UV_PVM)
IF (.NOT. L2D) THEN
  CALL GY_U_UV_DEVICE(PUM,PDYY,PDZZ,PDZY,GY_U_UV_PUM)
END IF
#endif
!
!
!*      12.   < U'V'>
!             -------
!
!
#ifndef MNH_OPENACC
IF (.NOT. L2D) THEN
  ZFLX(:,:,:)= - XCMFS * MYM(MXM(PK)) * (GY_U_UV_PUM + GX_V_UV_PVM)
ELSE
  ZFLX(:,:,:)= - XCMFS * MYM(MXM(PK)) * (GX_V_UV_PVM)
END IF
#else
CALL MXM_DEVICE(PK,ZTMP1_DEVICE)
CALL MYM_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
IF (.NOT. L2D) THEN
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
    ZFLX(JI,JJ,JK)= - XCMFS * ZTMP2_DEVICE(JI,JJ,JK) * (GY_U_UV_PUM(JI,JJ,JK) + GX_V_UV_PVM(JI,JJ,JK))
  END DO !CONCURRENT   
  !$acc end kernels
ELSE
  !$acc kernels present_cr(ZFLX)
  ZFLX(:,:,:)= - XCMFS * ZTMP2_DEVICE * (GX_V_UV_PVM)
  !$acc end kernels
END IF
#endif
!
!$acc kernels present_cr(ZFLX)
ZFLX(:,:,IKE+1)= ZFLX(:,:,IKE)
!$acc end kernels
!
!
! Compute the correlation at the first physical level with the following 
! hypothesis du/dz vary in 1/z and w=0 at the ground
#ifndef MNH_OPENACC
ZFLX(:,:,IKB:IKB)   = - XCMFS * MYM(MXM(PK(:,:,IKB:IKB))) *  (     &
  ( DYM( PUM(:,:,IKB:IKB) )                                        &
   -MYM( (PUM(:,:,IKB+1:IKB+1)-PUM(:,:,IKB:IKB))                   &
        *(1./MXM(PDZZ(:,:,IKB+1:IKB+1))+1./MXM(PDZZ(:,:,IKB:IKB))))&
    *0.5*MXM((PDZY(:,:,IKB+1:IKB+1)+PDZY(:,:,IKB:IKB)))            &
  ) / MXM(PDYY(:,:,IKB:IKB))                                       &
 +( DXM( PVM(:,:,IKB:IKB) )                                        &
   -MXM( (PVM(:,:,IKB+1:IKB+1)-PVM(:,:,IKB:IKB))                   &
        *(1./MYM(PDZZ(:,:,IKB+1:IKB+1))+1./MYM(PDZZ(:,:,IKB:IKB))))&
    *0.5*MYM((PDZX(:,:,IKB+1:IKB+1)+PDZX(:,:,IKB:IKB)))            &
  ) / MYM(PDXX(:,:,IKB:IKB))                                 )
#else
CALL DYM_DEVICE(PUM(:,:,IKB:IKB),ZTMP1_DEVICE(:,:,1:1))
!
CALL MXM_DEVICE(PDZZ(:,:,IKB+1:IKB+1),ZTMP2_DEVICE(:,:,1:1))
CALL MXM_DEVICE(PDZZ(:,:,IKB:IKB),ZTMP3_DEVICE(:,:,1:1))
!
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP5_DEVICE)
#else
!$acc kernels 
#endif
ZTMP5_DEVICE(:,:,1) = (PUM(:,:,IKB+1)-PUM(:,:,IKB))*(1./ZTMP2_DEVICE(:,:,1)+1./ZTMP3_DEVICE(:,:,1))
!$acc end kernels
CALL MYM_DEVICE(ZTMP5_DEVICE(:,:,1:1),ZTMP4_DEVICE(:,:,1:1))
!
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP5_DEVICE)
#else
!$acc kernels 
#endif
ZTMP5_DEVICE(:,:,1) = PDZY(:,:,IKB+1) + PDZY(:,:,IKB)
!$acc end kernels
CALL MXM_DEVICE(ZTMP5_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
!
CALL MXM_DEVICE(PDYY(:,:,IKB:IKB),ZTMP3_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP5_DEVICE)
#else
!$acc kernels 
#endif
ZTMP5_DEVICE(:,:,1) = ( ZTMP1_DEVICE(:,:,1) - ZTMP4_DEVICE(:,:,1)*0.5*ZTMP2_DEVICE(:,:,1) ) &
                      / ZTMP3_DEVICE(:,:,1)
!$acc end kernels
!
!
CALL DXM_DEVICE(PVM(:,:,IKB:IKB),ZTMP1_DEVICE(:,:,1:1))
!
CALL MYM_DEVICE(PDZZ(:,:,IKB:IKB),ZTMP2_DEVICE(:,:,1:1))
CALL MYM_DEVICE(PDZZ(:,:,IKB+1:IKB+1),ZTMP3_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP6_DEVICE)
#else
!$acc kernels 
#endif
ZTMP6_DEVICE(:,:,1) = (PVM(:,:,IKB+1)-PVM(:,:,IKB))*(1./ZTMP3_DEVICE(:,:,1)+1./ZTMP2_DEVICE(:,:,1))
!$acc end kernels
CALL MXM_DEVICE(ZTMP6_DEVICE(:,:,1:1),ZTMP4_DEVICE(:,:,1:1))
!
!$acc kernels present_cr(ZTMP6_DEVICE)
ZTMP6_DEVICE(:,:,1:1) = (PDZX(:,:,IKB+1:IKB+1)+PDZX(:,:,IKB:IKB))
!$acc end kernels
CALL MYM_DEVICE(ZTMP6_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
!
CALL MYM_DEVICE(PDXX(:,:,IKB:IKB),ZTMP3_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP6_DEVICE)
#else
!$acc kernels 
#endif
ZTMP6_DEVICE(:,:,1) = ( ZTMP1_DEVICE(:,:,1) - ZTMP4_DEVICE(:,:,1)*0.5*ZTMP2_DEVICE(:,:,1) ) &
                      / ZTMP3_DEVICE(:,:,1)
!$acc end kernels
!
CALL MXM_DEVICE(PK(:,:,IKB:IKB),ZTMP1_DEVICE(:,:,1:1))
CALL MYM_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
!
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZFLX)
#else
!$acc kernels 
#endif
ZFLX(:,:,IKB) = - XCMFS * ZTMP2_DEVICE(:,:,1) * ( ZTMP5_DEVICE(:,:,1) + ZTMP6_DEVICE(:,:,1) )
!$acc end kernels
#endif
! 
! extrapolates this flux under the ground with the surface flux
!$acc parallel present_cr(ZFLX)
#ifndef MNH_BITREP
!$mnh_expand_array(JI=1:JIU,JJ=1:JJU)
 ZFLX(:,:,IKB-1) =                                                           &
   PTAU11M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) * PDIRCOSZW(:,:)**2         &
  +PTAU12M(:,:) * (PCOSSLOPE(:,:)**2 - PSINSLOPE(:,:)**2) *                   &
                  PDIRCOSZW(:,:)**2                                           &
  -PTAU22M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:)                             &
  +PTAU33M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) * ZDIRSINZW(:,:)**2         &
  -PCDUEFF(:,:) * (                                                           &
    2. * PUSLOPEM(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) *                    &
          PDIRCOSZW(:,:) * ZDIRSINZW(:,:)                                     &
    +PVSLOPEM(:,:) * (PCOSSLOPE(:,:)**2 - PSINSLOPE(:,:)**2) * ZDIRSINZW(:,:) )
!$mnh_end_expand_array()
#else
!PW: BUG: commented 'acc loop independent collapse(2)' to workaround compiler bug (NVHPC 21.1)
!$mnh_expand_array(JI=1:JIU,JJ=1:JJU)
 ZFLX(:,:,IKB-1) =                                                           &
   PTAU11M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) * BR_P2(PDIRCOSZW(:,:))     &
  +PTAU12M(:,:) * (BR_P2(PCOSSLOPE(:,:)) - BR_P2(PSINSLOPE(:,:))) *           &
                  BR_P2(PDIRCOSZW(:,:))                                       &
  -PTAU22M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:)                             &
  +PTAU33M(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) * BR_P2(ZDIRSINZW(:,:))     &
  -PCDUEFF(:,:) * (                                                           &
    2. * PUSLOPEM(:,:) * PCOSSLOPE(:,:) * PSINSLOPE(:,:) *                    &
          PDIRCOSZW(:,:) * ZDIRSINZW(:,:)                                     &
          +PVSLOPEM(:,:) * (BR_P2(PCOSSLOPE(:,:)) - BR_P2(PSINSLOPE(:,:))) * ZDIRSINZW(:,:) )
!$mnh_end_expand_array()
#endif
!$acc end parallel
!  
#ifndef MNH_OPENACC
ZFLX(:,:,IKB-1:IKB-1) = 2. * MXM( MYM( ZFLX(:,:,IKB-1:IKB-1) ) )  &
                   - ZFLX(:,:,IKB:IKB)
#else
CALL MYM_DEVICE(ZFLX(:,:,IKB-1:IKB-1),ZTMP1_DEVICE(:,:,1:1))
CALL MXM_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZFLX)
#else
!$acc kernels 
#endif
ZFLX(:,:,IKB-1) = 2. * ZTMP2_DEVICE(:,:,1) - ZFLX(:,:,IKB)
!$acc end kernels
#endif
!     
! stores  <U V>
IF ( tpfile%lopened .AND. OTURB_FLX ) THEN
  TZFIELD%CMNHNAME   = 'UV_FLX'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'UV_FLX'
  TZFIELD%CUNITS     = 'm2 s-2'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'X_Y_Z_UV_FLX'
  TZFIELD%NGRID      = 5
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLX)
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLX)
END IF
!
!
!
!computation of the source for rho*V due to this flux
#ifndef MNH_OPENACC
IF (.NOT. LFLAT) THEN
  PRUS(:,:,:) = PRUS(:,:,:)                                &
              - DYF(ZFLX * MXM(MYM(PRHODJ) * PINV_PDYY) )         &
              + DZF( MYF( MZM(ZFLX)*MXM(PDZY/MZM(PDYY)))   &
                    * MXM(PMZM_PRHODJ * PINV_PDZZ) )
ELSE
  PRUS(:,:,:) = PRUS(:,:,:) - DYF(ZFLX * MXM(MYM(PRHODJ) * PINV_PDYY) )
END IF
#else
CALL MYM_DEVICE(PRHODJ,ZTMP1_DEVICE)
!$acc kernels
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(3)
#endif
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * PINV_PDYY(JI,JJ,JK)
END DO !CONCURRENT   
!$acc end kernels
CALL MXM_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE) 
!$acc kernels
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(3)
#endif
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = ZFLX(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
END DO !CONCURRENT   
!$acc end kernels
CALL DYF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
IF (.NOT. LFLAT) THEN
  CALL MZM_DEVICE(ZFLX,ZTMP2_DEVICE)
  CALL MZM_DEVICE(PDYY,ZTMP3_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP4_DEVICE(JI,JJ,JK) = PDZY(JI,JJ,JK)/ZTMP3_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MXM_DEVICE(ZTMP4_DEVICE,ZTMP5_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP4_DEVICE(JI,JJ,JK) = ZTMP2_DEVICE(JI,JJ,JK)*ZTMP5_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MYF_DEVICE(ZTMP4_DEVICE,ZTMP2_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP3_DEVICE(JI,JJ,JK) = PMZM_PRHODJ(JI,JJ,JK) * PINV_PDZZ(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MXM_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP5_DEVICE(JI,JJ,JK) = ZTMP2_DEVICE(JI,JJ,JK)*ZTMP4_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT    
  !$acc end kernels
  CALL DZF_DEVICE( ZTMP5_DEVICE, ZTMP3_DEVICE )
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     PRUS(JI,JJ,JK) = PRUS(JI,JJ,JK) - ZTMP1_DEVICE(JI,JJ,JK) + ZTMP3_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
ELSE
!$acc kernels present_cr(PRUS)
  PRUS(:,:,:) = PRUS(:,:,:) - ZTMP1_DEVICE
!$acc end kernels
END IF
#endif
!
!computation of the source for rho*V due to this flux
#ifndef MNH_OPENACC
IF (.NOT. LFLAT) THEN
  PRVS(:,:,:) = PRVS(:,:,:)                             &
                - DXF(ZFLX * MYM(MXM(PRHODJ) * PINV_PDXX) )    &
                + DZF( MXF( MZM(ZFLX)*MYM(PDZX/MZM(PDXX))) &
                      * MYM(PMZM_PRHODJ * PINV_PDZZ) )
ELSE
  PRVS(:,:,:) = PRVS(:,:,:) - DXF(ZFLX * MYM(MXM(PRHODJ) * PINV_PDXX) )
END IF
#else
CALL MXM_DEVICE(PRHODJ,ZTMP1_DEVICE)
!$acc kernels
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(3)
#endif
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * PINV_PDXX(JI,JJ,JK)
END DO !CONCURRENT   
!$acc end kernels
CALL MYM_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE) 
!$acc kernels
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(3)
#endif
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = ZFLX(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
END DO !CONCURRENT   
!$acc end kernels
CALL DXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
IF (.NOT. LFLAT) THEN
  CALL MZM_DEVICE(ZFLX,ZTMP2_DEVICE)
  CALL MZM_DEVICE(PDXX,ZTMP3_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP4_DEVICE(JI,JJ,JK) = PDZX(JI,JJ,JK)/ZTMP3_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MYM_DEVICE(ZTMP4_DEVICE,ZTMP5_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP4_DEVICE(JI,JJ,JK) = ZTMP2_DEVICE(JI,JJ,JK)*ZTMP5_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MXF_DEVICE(ZTMP4_DEVICE,ZTMP2_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP3_DEVICE(JI,JJ,JK) = PMZM_PRHODJ(JI,JJ,JK) * PINV_PDZZ(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL MYM_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP5_DEVICE(JI,JJ,JK) = ZTMP2_DEVICE(JI,JJ,JK)*ZTMP4_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL DZF_DEVICE( ZTMP5_DEVICE, ZTMP3_DEVICE )
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     PRVS(JI,JJ,JK) = PRVS(JI,JJ,JK) - ZTMP1_DEVICE(JI,JJ,JK) + ZTMP3_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
ELSE
!$acc kernels
  PRVS(:,:,:) = PRVS(:,:,:) - ZTMP1_DEVICE
!$acc end kernels
END IF
#endif
!
IF (KSPLT==1) THEN
  !
  !Contribution to the dynamic production of TKE:
  !
#ifndef MNH_OPENACC
  IF (.NOT. L2D) THEN
    ZWORK(:,:,:) = - MXF( MYF( ZFLX *                                &
       (GY_U_UV_PUM + GX_V_UV_PVM) ) ) 
  ELSE
    ZWORK(:,:,:) = - MXF( MYF( ZFLX *                                &
       (GX_V_UV_PVM) ) )  
  ENDIF
#else
  IF (.NOT. L2D) THEN
     !$acc kernels
#ifdef MNH_COMPILER_NVHPC
     !$acc loop independent collapse(3)
#endif
     DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
        ZTMP1_DEVICE(JI,JJ,JK) = ZFLX(JI,JJ,JK) * (GY_U_UV_PUM(JI,JJ,JK) + GX_V_UV_PVM(JI,JJ,JK))
     END DO !CONCURRENT   
!$acc end kernels
  ELSE
!$acc kernels present_cr(ZTMP1_DEVICE)
    ZTMP1_DEVICE = ZFLX * GX_V_UV_PVM
!$acc end kernels
  ENDIF
  CALL MYF_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
  CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZWORK(JI,JJ,JK) = - ZTMP1_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
#endif
  !
  ! evaluate the dynamic production at w(IKB+1) in PDP(IKB)
  !
#ifndef MNH_OPENACC
  ZWORK(:,:,IKB:IKB) =  -                                             &
     MXF ( MYF( 0.5 * (ZFLX(:,:,IKB+1:IKB+1)+ZFLX(:,:,IKB:IKB)) ) )   &
   *(MXF ( MYF(                                                       &
       DYM( 0.5 * (PUM(:,:,IKB+1:IKB+1)+PUM(:,:,IKB:IKB))  )          &
      / MXM( 0.5*(PDYY(:,:,IKB:IKB)+PDYY(:,:,IKB+1:IKB+1)) )          &
      +DXM( 0.5 * (PVM(:,:,IKB+1:IKB+1)+PVM(:,:,IKB:IKB))  )          &
      / MYM( 0.5*(PDXX(:,:,IKB:IKB)+PDXX(:,:,IKB+1:IKB+1)) )          &
         )    )                                                       &  
    -MXF( (PUM(:,:,IKB+1:IKB+1)-PUM(:,:,IKB:IKB)) /                   &
              MXM(PDZZ(:,:,IKB+1:IKB+1)) * PDZY(:,:,IKB+1:IKB+1)      &
        ) / MYF(MXM( 0.5*(PDYY(:,:,IKB:IKB)+PDYY(:,:,IKB+1:IKB+1)) ) )&
    -MYF( (PVM(:,:,IKB+1:IKB+1)-PVM(:,:,IKB:IKB)) /                   &
              MYM(PDZZ(:,:,IKB+1:IKB+1)) * PDZX(:,:,IKB+1:IKB+1)      &
        ) / MXF(MYM( 0.5*(PDXX(:,:,IKB:IKB)+PDXX(:,:,IKB+1:IKB+1)) ) )&
    ) 
#else
#ifdef MNH_COMPILER_CCE
  !$acc kernels present(ZTMP1_DEVICE)
#else
  !$acc kernels 
#endif
  ZTMP1_DEVICE(:,:,1) = 0.5 * (ZFLX(:,:,IKB+1)+ZFLX(:,:,IKB))
  !$acc end kernels
  CALL MYF_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
  CALL MXF_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP1_DEVICE(:,:,1:1))
  !
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = 0.5 * (PUM(:,:,IKB+1)+PUM(:,:,IKB))
!$acc end kernels
  CALL DYM_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP3_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = 0.5*(PDYY(:,:,IKB)+PDYY(:,:,IKB+1))
!$acc end kernels
  CALL MXM_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP4_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = 0.5 * (PVM(:,:,IKB+1)+PVM(:,:,IKB))
!$acc end kernels
  CALL DXM_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP5_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = 0.5*(PDXX(:,:,IKB)+PDXX(:,:,IKB+1))
!$acc end kernels
  CALL MYM_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP6_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP2_DEVICE)
#else
!$acc kernels 
#endif
  ZTMP2_DEVICE(:,:,1) = ZTMP3_DEVICE(:,:,1) / ZTMP4_DEVICE(:,:,1) + ZTMP5_DEVICE(:,:,1) / ZTMP6_DEVICE(:,:,1)
!$acc end kernels
  CALL MYF_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP3_DEVICE(:,:,1:1))
  CALL MXF_DEVICE(ZTMP3_DEVICE(:,:,1:1),ZTMP2_DEVICE(:,:,1:1))
  !
  CALL MXM_DEVICE(PDZZ(:,:,IKB+1:IKB+1),ZTMP3_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP5_DEVICE)
#else
!$acc kernels 
#endif
  ZTMP5_DEVICE(:,:,1) = (PUM(:,:,IKB+1)-PUM(:,:,IKB)) / ZTMP3_DEVICE(:,:,1) * PDZY(:,:,IKB+1)
!$acc end kernels
  CALL MXF_DEVICE(ZTMP5_DEVICE(:,:,1:1),ZTMP3_DEVICE(:,:,1:1))
  CALL MYF_DEVICE(ZTMP4_DEVICE(:,:,1:1),ZTMP5_DEVICE(:,:,1:1)) !Re-use of ZTMP4_DEVICE
  !
  CALL MYM_DEVICE(PDZZ(:,:,IKB+1:IKB+1),ZTMP4_DEVICE(:,:,1:1))
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZTMP7_DEVICE)
#else
!$acc kernels 
#endif
  ZTMP7_DEVICE(:,:,1) = (PVM(:,:,IKB+1)-PVM(:,:,IKB)) / ZTMP4_DEVICE(:,:,1) * PDZX(:,:,IKB+1)
!$acc end kernels
  CALL MYF_DEVICE(ZTMP7_DEVICE(:,:,1:1),ZTMP4_DEVICE(:,:,1:1))
  CALL MXF_DEVICE(ZTMP6_DEVICE(:,:,1:1),ZTMP7_DEVICE(:,:,1:1)) !Re-use of ZTMP6_DEVICE
  !  
#ifdef MNH_COMPILER_CCE
!$acc kernels present(ZWORK)
#else
!$acc kernels 
#endif
  ZWORK(:,:,IKB) =  - ZTMP1_DEVICE(:,:,1) * (ZTMP2_DEVICE(:,:,1) &
                                             - ZTMP3_DEVICE(:,:,1) / ZTMP5_DEVICE(:,:,1) &
                                             - ZTMP4_DEVICE(:,:,1) / ZTMP7_DEVICE(:,:,1) ) 
!$acc end kernels
#endif
  !
  ! dynamic production 
  !$acc kernels present_cr(PDP)
  PDP(:,:,:) = PDP(:,:,:) + ZWORK(:,:,:)
  !$acc end kernels
  ! 
END IF
!
! Storage in the LES configuration
!
IF (LLES_CALL .AND. KSPLT==1) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
  CALL LES_MEAN_SUBGRID( MXF(MYF(ZFLX)), X_LES_SUBGRID_UV ) 
  CALL LES_MEAN_SUBGRID( MXF(MYF(GY_U_UV(PUM,PDYY,PDZZ,PDZY)*ZFLX)), X_LES_RES_ddxa_U_SBG_UaU , .TRUE.)
  CALL LES_MEAN_SUBGRID( MXF(MYF(GX_V_UV(PVM,PDXX,PDZZ,PDZX)*ZFLX)), X_LES_RES_ddxa_V_SBG_UaV , .TRUE.)
#else
!$acc data copy(X_LES_SUBGRID_UV,X_LES_RES_ddxa_U_SBG_UaU,X_LES_RES_ddxa_V_SBG_UaV)
  CALL MYF_DEVICE(ZFLX,ZTMP1_DEVICE)
  CALL MXF_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
  CALL LES_MEAN_SUBGRID(ZTMP2_DEVICE,X_LES_SUBGRID_UV)
  !
  CALL GY_U_UV_DEVICE(PUM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
!$acc kernels present_cr(ZTMP3_DEVICE)
  ZTMP3_DEVICE = ZTMP1_DEVICE*ZFLX
!$acc end kernels
  CALL MYF_DEVICE(ZTMP3_DEVICE,ZTMP2_DEVICE)
  CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL LES_MEAN_SUBGRID(ZTMP1_DEVICE, X_LES_RES_ddxa_U_SBG_UaU , .TRUE.)
  !
  CALL GX_V_UV_DEVICE(PVM,PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
!$acc kernels present_cr(ZTMP3_DEVICE)
  ZTMP3_DEVICE = ZTMP1_DEVICE*ZFLX
!$acc end kernels
  CALL MYF_DEVICE(ZTMP3_DEVICE,ZTMP2_DEVICE)
  CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL LES_MEAN_SUBGRID(ZTMP1_DEVICE, X_LES_RES_ddxa_V_SBG_UaV , .TRUE.)
!$acc end data
#endif
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF

if ( mppdb_initialized ) then
  !Check all inout arrays
  call Mppdb_check( prus, "Turb_hor_uv end:prus" )
  call Mppdb_check( prvs, "Turb_hor_uv end:prvs" )
  call Mppdb_check( pdp,  "Turb_hor_uv end:pdp"  )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate ( ZFLX, ZWORK, ZDIRSINZW, GY_U_UV_PUM, GX_V_UV_PVM )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif
     

!$acc end data

END SUBROUTINE TURB_HOR_UV
